"""Exceptions."""


class WebsaunaConfigException(Exception):
    """Websauna config exception base class."""

    pass


class InvalidResourceScheme(WebsaunaConfigException):
    """Include reference does not exist."""

    pass


class NonExistingInclude(WebsaunaConfigException):
    """Include reference does not exist."""

    pass



"""Support for INI file inclusion mechanism."""
# Standard Library
import logging
import os
import typing as t
from logging.config import fileConfig

import plaster_pastedeploy
from paste.deploy import loadwsgi

# Websauna
from websauna.utils.config.includer import IncludeAwareConfigParser


class ConfigLoader(loadwsgi.ConfigLoader):
    """Configuration Loader."""

    def __init__(self, filename: str):
        """Initialize the config loader."""
        self.filename = filename = filename.strip()
        defaults = {
            'here': os.path.dirname(os.path.abspath(filename)),
            '__file__': os.path.abspath(filename)
        }
        self.parser = IncludeAwareConfigParser(filename, defaults=defaults)
        with open(filename) as f:
            self.parser.read_file(f)


class Loader(plaster_pastedeploy.Loader):
    """Loader returning a WSGI application."""

    def __init__(self, uri):
        """Initialize the Loader."""
        # Replace our scheme for file, so it could be handled as usual.
        uri.scheme = 'file'
        super().__init__(uri)

    def _get_loader(self, defaults: t.Optional[dict]=None) -> ConfigLoader:
        """Return a ConfigLoader instance.

        :param defaults: Dict with default values.
        :return: ConfigLoader instance.
        """
        defaults = self._get_defaults(defaults)
        loader = ConfigLoader(self.uri.path)
        loader.update_defaults(defaults)
        return loader

    def _get_parser(self, defaults: t.Optional[dict]=None) -> IncludeAwareConfigParser:
        """Return an instance of IncludeAwareConfigParser.

        :param defaults: Dict with default values.
        :return: IncludeAwareConfigParser instance.
        """
        return self._get_loader(defaults).parser

    def setup_logging(self, defaults: t.Optional[dict]=None, disable_existing_loggers=False):
        """Set up logging via :func:`logging.config.fileConfig`.

        Defaults are specified for the special ``__file__`` and ``here``
        variables, similar to PasteDeploy config loading. Extra defaults can
        optionally be specified as a dict in ``defaults``.

        :param defaults: The defaults that will be used when passed to :func:`logging.config.fileConfig`.
        :param disable_existing_loggers: Should existing loggers be disabled.
        """
        defaults = defaults if defaults else {}
        if 'loggers' in self.get_sections():
            uri_path = self.uri.path
            parser = self._get_parser()
            defaults.update(
                {
                    '__file__': uri_path,
                    'here': os.path.dirname(uri_path),
                    'disable_existing_loggers': disable_existing_loggers
                }
            )
            fileConfig(
                parser,
                defaults,
            )
        else:
            logging.basicConfig()

    def get_wsgi_app(self, name=None, defaults=None):
        return self._get_loader(defaults).get_app(name, defaults)

    def get_wsgi_filter(self, name=None, defaults=None):
        return self._get_loader(defaults).get_filter(name, defaults)

    def get_wsgi_server(self, name=None, defaults=None):
        return self._get_loader(defaults).get_server(name, defaults)

    def __repr__(self) -> str:
        """Representation of this object.

        :return: String representation of this object.
        """
        klass_module = self.__class__.__module__
        klass_qualname = self.__class__.__qualname__
        return '{0}.{1}(uri="{2}")'.format(klass_module, klass_qualname, self.uri)



"""Test utils.traverseattribute."""
# Websauna
from websauna.utils import traverseattribute


class Dummy:
    """Dummy class to be traversed."""
    __parent__ = None


def test_traverse_attribute():
    """Test traverseattribute.traverse_attribute."""
    func = traverseattribute.traverse_attribute
    grandparent = Dummy()
    parent = Dummy()
    parent.__parent__ = grandparent
    obj = Dummy()
    obj.__parent__ = parent
    tree = list(func(obj, '__parent__'))

    assert len(tree) == 3
    assert tree[0] == obj
    assert tree[1] == parent
    assert tree[2] == grandparent



"""Test utils.time."""
# Standard Library
from datetime import datetime
from datetime import timezone

# Websauna
from websauna.utils import time


def test_now():
    """Test time.now."""
    func = time.now
    result = func()

    assert isinstance(result, datetime)
    assert result.tzinfo == timezone.utc



"""Support for INI file inclusion mechanism."""
# Standard Library
import configparser
import io
import os
import typing as t
from urllib.parse import urlparse

# Pyramid
from pyramid.settings import aslist

import pkg_resources
from paste.deploy import loadwsgi

# Websauna
from websauna.utils.config import exceptions as exc
from websauna.utils.config import _resource_manager


_VALID_SCHEMAS_ = (
    'resource',
)


class IncludeAwareConfigParser(loadwsgi.NicerConfigParser):
    """Include .ini settings file in others.

    This is a hack to get quick include support for Pyramid INI settings files. This is a variation of Python ``ConfigParser`` which knows about ``[includes]`` section. In this section you can tell the config file to include other config files.

    To add includes add a section in your INI::

        [includes]
        include_ini_files =
            resource://websauna/conf/production.ini
            resource://websauna/conf/base.ini


    Each included file is referred by URL. Currently support protocols are:

    * ``resource:``:  This scheme indicates a Python resource specification. The includes are provided `as Python packages resources <https://pythonhosted.org/setuptools/pkg_resources.html>`_.

    The keys in the current INI file sections are added from includes if the keys do not exist yet. Includes are processed from first to last, first taking precedences.

    The include is not recursive. E.g. the includes of include are not processed at the moment.
    """

    optionxform = str

    def _read(self, fp: io.TextIOWrapper, fpname: str):
        """Read configuration file and process its includes.

        :param fp: TextIOWrapper
        :param fpname: Main configuration filename.
        """
        super()._read(fp, fpname)
        self.process_includes(fpname)

    def resolve(self, include_file: str, fpname: str) -> t.TextIO:
        """Resolve include_file and return a readable file like object.

        :param include_file: File to be include.
        :param fpname: Main configuration filename.
        :return: Return a readable file-like object for the specified resource.
        """
        parts = urlparse(include_file)
        if parts.scheme not in _VALID_SCHEMAS_:
            raise exc.InvalidResourceScheme(
                "Supported resources: {resources}. Got {include} in {fpname}".format(
                    resources=', '.join(_VALID_SCHEMAS_),
                    include=include_file,
                    fpname=fpname
                )
            )

        package = parts.netloc
        args = package.split('.') + [parts.path.lstrip('/')]
        path = os.path.join(*args)

        req = pkg_resources.Requirement.parse(package)

        if not _resource_manager.resource_exists(req, path):
            raise exc.NonExistingInclude(
                "Could not find {include}".format(include=include_file)
            )

        config_source = _resource_manager.resource_stream(req, path)
        return config_source

    def read_include(self, include_file: str, fpname: str):
        """Augment the current config entries from another INI file.

        :param include_file: File to be include.
        :param fpname: Main configuration filename.
        """
        fp = self.resolve(include_file, fpname)
        config = configparser.ConfigParser()

        # For some reason configparser refuses to work with file handles opened by pkg_resource
        text = fp.read().decode('utf-8')
        config.read_string(text, source=include_file)

        for s in config.sections():
            source_section = s
            target_section = s

            if target_section not in self.sections():
                self.add_section(target_section)

            # What we have currently - include must not override existing settings
            current_settings = [key for key, value in self.items(target_section, raw=True)]

            # Go through included settings and add them if the setting is not yet there
            for key, value in config.items(source_section, raw=True):
                if key not in current_settings:
                    self._sections[target_section][key] = value

    def process_includes(self, fpname: str):
        """Process includes section.

        :param fpname: Main configuration filename.
        """
        if 'includes' in self.sections():
            include_ini_files = aslist(self.get('includes', 'include_ini_files'))
            for include in include_ini_files:
                try:
                    self.read_include(include, fpname)
                except pkg_resources.DistributionNotFound as not_found_error:
                    if '/websauna/' in include and '/websauna/system/' not in include:
                        include = include.replace('/websauna/', '/websauna.system/')
                        try:
                            self.read_include(include, fpname)
                        except pkg_resources.DistributionNotFound:
                            raise not_found_error
                    else:
                        raise

    @classmethod
    def retrofit_settings(cls, global_config: dict, section: str='app:main') -> dict:
        """Update settings dictionary given to WSGI application constructor by Paster to have included settings.

        :param global_config: global_config dict as given by Paster
        :param section: Section to be processed.
        :return: New instance of settings
        """
        assert global_config
        assert '__file__' in global_config

        parser = cls()
        parser.read(global_config["__file__"])
        return {k: v for k, v in parser.items(section)}



"""Test utils.jsonb."""
# Standard Library
from unittest import mock

import pytest

from websauna.utils import jsonb


@mock.patch("websauna.utils.jsonb._sqlalchemy_present", False)
def test_is_index_property_raises_on_no_sqlalchemy():
    """On no sqlalchemy, calling `is_index_property` should raise RuntimeError."""
    with pytest.raises(RuntimeError):
        jsonb.is_index_property(obj=object(), name="")


class MockIndexProperty:
    pass


class MockObj:
    mock_property = MockIndexProperty()
    other_property = object()


@mock.patch("websauna.utils.jsonb._sqlalchemy_present", True)
def test_is_index_property_works():
    """If sqlalchemy is present,  `is_index_property` should identify it properly."""
    if not hasattr(jsonb, "index_property"):
        jsonb.index_property = None
    with mock.patch("websauna.utils.jsonb.index_property", MockIndexProperty):
        assert jsonb.is_index_property(obj=MockObj(), name="mock_property")
        assert not jsonb.is_index_property(obj=MockObj(), name="other_property")



"""Time and date helpers."""
# Standard Library
import datetime


def now() -> datetime.datetime:
    """Get the current time as timezone-aware UTC timestamp.

    :return: Current datetime with UTC timezone
    """
    return datetime.datetime.now(datetime.timezone.utc)



"""Cryptographic utilities."""
# Standard Library
import random
import string


_default = string.ascii_lowercase + string.ascii_uppercase + string.digits


def generate_random_string(length: int, letters: str=_default) -> str:
    """Generate cryptographically safe random string.

    :param length: String length
    :param letters: Choose from this letter pool
    """
    return ''.join(random.SystemRandom().choice(letters) for _ in range(length))



"""HTML manipulation helpers."""
# Standard Library
import typing as t


_js_escapes = {
    ord('\\'): '\\u005C',
    ord('\''): '\\u0027',
    ord('"'): '\\u0022',
    ord('>'): '\\u003E',
    ord('<'): '\\u003C',
    ord('&'): '\\u0026',
    ord('='): '\\u003D',
    ord('-'): '\\u002D',
    ord(';'): '\\u003B',
    ord('\u2028'): '\\u2028',
    ord('\u2029'): '\\u2029'
}

# Escape every ASCII character with a value less than 32.
_js_escapes.update((z, '\\u%04X' % z) for z in range(32))


def escape_js(value: t.AnyStr) -> str:
    """Hex encodes characters for use in JavaScript strings.

    :param value: String to be escaped.
    :return: A string safe to be included inside a <script> tag.
    """
    return str(value).translate(_js_escapes)



"""Test utils.qualname."""
import pytest

# Websauna
from websauna.utils import html
from websauna.utils import qualname
from websauna.utils import slug
from websauna.utils import time


test_data = (
    (html.escape_js, 'websauna.utils.html.escape_js'),
    (qualname.get_qual_name, 'websauna.utils.qualname.get_qual_name'),
    (slug.slug_to_uuid, 'websauna.utils.slug.slug_to_uuid'),
    (slug.uuid_to_slug, 'websauna.utils.slug.uuid_to_slug'),
    (time.now, 'websauna.utils.time.now'),
)


@pytest.mark.parametrize('value,expected', test_data)
def test_get_qual_name(value, expected):
    """Test qualname.get_qual_name."""
    func = qualname.get_qual_name
    assert func(value) == expected



"""Test utils.slug."""
# Standard Library
import uuid

# Websauna
from websauna.utils.slug import slug_to_uuid
from websauna.utils.slug import uuid_to_slug


def test_uuid_slug():
    """Make short slugs for 64-bit UUID ids."""
    _uuid = uuid.uuid4()

    _uuid2 = slug_to_uuid(uuid_to_slug(_uuid))
    assert _uuid == _uuid2



"""Utilities to encode and decode UUID objects as URL parameters.

We provide UUID to Base64 encoded string and back coding.

`More information what term slug means <http://stackoverflow.com/questions/427102/what-is-a-slug-in-django>`_.
"""
# Standard Library
import base64
import binascii
import uuid


class SlugDecodeError(Exception):
    """Raised when you pass invalid b64 slug data."""


def uuid_to_slug(uuid_: uuid.UUID) -> str:
    """Convert UUID object to a compact base64 string presentation.

    .. note ::

        Slugs are supposed to be human readable. We are stretching that definition here a bit.

    :param uuid_: UUID object

    :return: String like ''I0p4RyoIQe-EQ1GU_QicoQ'
    """

    # Catch some common typing errors
    assert isinstance(uuid_, uuid.UUID)

    encoded = base64.b64encode(uuid_.bytes)

    # https://docs.python.org/2/library/base64.html#base64.urlsafe_b64encode

    # URLs don't like +
    return encoded.decode("utf-8").rstrip('=\n').replace('/', '_').replace("+", "-")


def slug_to_uuid(slug: str) -> uuid.UUID:
    """Convert UUID URL slug string to UUID object.

    This method raises `ValueError` if slug cannot be converted to UUID:

    .. code-block:: python

        try:
            # Is like 'I0p4RyoIQe-EQ1GU_QicoQ'
            delivery_uuid = slug_to_uuid(art_data["cartID"])
        except SlugDecodeError as e:
            return HTTPNotFound("UUID was not a well-formed base64 encoded slug") from e

    :param slug: Base64 string presentation of UUID

    :raise: SlugDecodeError if the is an issue to decode slug
    """

    assert type(slug) == str

    try:
        bytes = (slug + '==').replace('_', '/').replace("-", "+")
        bytes = base64.b64decode(bytes)
        return uuid.UUID(bytes=bytes)
    except (ValueError, binascii.Error) as e:
        raise SlugDecodeError("Cannot decode supposed B64 slug: {}".format(slug)) from e



"""Test utils.crypt."""
# Standard Library
import string

import pytest

# Websauna
from websauna.utils import crypt


test_data = (
    (32, None, crypt._default),
    (32, crypt._default, crypt._default),
    (40, string.digits, string.digits),
)


@pytest.mark.parametrize('length,letters,pool', test_data)
def test_generate_random_string(length, letters, pool):
    """Test crypt.generate_random_string.

    Test if length and letters params are respected.
    """
    func = crypt.generate_random_string
    params = {
        'length': length,
    }
    if letters:
        params['letters'] = letters
    result = func(**params)
    assert len(result) == length
    assert len([c for c in result if c in pool]) == length



"""Automatically fire events on function enter and exit.

Use :py:func:`@event_source` to mark methods which fire events::

    class MyClass:

        # Target function
        @event_source
        def my_func(self):
            print("In myfunc")

        def run(self):
            self.my_func()

Create a class which can bind handlers to be called before and after event_source methods::

    class MyHandlers:

        # Aspect function
        @before(MyClass.my_func)
        def extra_logging(self):
            print("Entering my_func")

        # Aspect function
        @after(MyClass.my_func)
        def extra_logging(self):
            print("After my_func")


    # Nothing happens yet
    instance_a = MyClass()
    instance_a.run()
    # ... In myfunc

    # Aspects run on instiated targetters only
    my_handler = MyHandlers()

    instance_b = MyClass()
    bind_events(instance_b, my_handler)
    instance_b.run()
    # ...
    # Entering myfunc
    # In myfunc
    # After myfunc

This implementation

* It is not designed to be a general solution, but targets a very narrow use case only. But you are free to expand.

* Works only on class methods.

* Event sources are named. By default they get name from function name, but you can override this to avoid namespace conflict

* Event handlers are not called if the subclass of ``MyClass`` overrides the parent method and doesn't call ``super()``. In this case you need to call :py:func:`fire_advisor_event` manually from the overriden method.

"""
# Standard Library
import functools
import inspect
import typing as t
from collections import defaultdict
from enum import Enum


#: List of unbound methods that are marked as advisors
_advisor_methods = defaultdict(list)

#: Bound methods to be called when the join point is executed as [joint point name] -> (role, bound_method,)  mappings
_event_source_hooks = []


class AdvisorRole(Enum):
    """Mark the role when storing references to advisor methods in our internal registry."""

    #: This advisor is to be called before the function
    before = 1

    #: This advisor is to be called after the function
    after = 2


def fire_advisor_event(source: object, event_source_name: str, role: AdvisorRole, *args, **kwargs):
    """Call bound advisors for a join point."""

    advisor_mappings = getattr(source, "_advisor_mappings", None)
    if not advisor_mappings:
        # No advisors bound
        return

    advisors = advisor_mappings.get(event_source_name, [])

    for target_role, advisor in advisors:
        if target_role == role:
            advisor(*args, **kwargs)


def event_source(method: t.Callable, name: t.Optional[str]=None):
    """A decorator which makes the function act as a source of before and after call events.

    You can later subscribe to these event with :py:func:`before` and :py:func`after` decorators.

    :param method: Target class method

    :param: Name of for the join point. If not given use the function name.
    """

    # We must use function name instead of function pointer for the registry, because function object changes with unbound vs. bound Python class methods

    if not name:
        name = method.__name__

    @functools.wraps(method)
    def _inner(*args, **kwargs):
        _self = args[0]
        fire_advisor_event(_self, name, AdvisorRole.before)
        retval = method(*args, **kwargs)
        fire_advisor_event(_self, name, AdvisorRole.after)
        return retval

    assert name not in _event_source_hooks, "There already exist event_source with same name"

    _event_source_hooks.append(name)
    method._event_source_name = name

    return _inner


def after(target_event_source: t.Callable):
    """Call decorated function before target function.

    :param target_event_source: Target method decorated with @event_source
    """

    def _outer(advisor_method):

        wrapped = getattr(target_event_source, "__wrapped__", None)
        assert wrapped, "The target function {} was not decorated with @event_source decorator or any other decorator".format(target_event_source)

        name = wrapped._event_source_name

        _advisor_methods[advisor_method].append((AdvisorRole.after, name,))
        return advisor_method

    return _outer


def before(target_event_source: t.Callable):
    """Call decorated function before target function.

    :param target_event_source: Target method decorated with @event_source
    """

    def _outer(advisor_method):

        wrapped = getattr(target_event_source, "__wrapped__", None)
        assert wrapped, "The target function {} was not decorated with @event_source decorator or any other decorator".format(target_event_source)

        name = wrapped._event_source_name

        _advisor_methods[advisor_method].append((AdvisorRole.before, name,))
        return advisor_method

    return _outer


def bind_events(source: object, target: object):
    """Making the advisor functions of instance active.

    To correctly handle application flow, before and after advices are not bind their target functions until the instance of a target class is contructed. This avoids import time side effects that advices would be always unconditionally bound.

    :param source: Target instance providing @event_source methods

    :param target: Provider of advisor methods
    """
    cls = target.__class__  # noQA

    # Store bound advisors on the object using a private attribute
    mappings = getattr(source, "_advisor_mappings", None)
    if not mappings:
        source._advisor_mappings = mappings = defaultdict(list)

    for name, bound_advisor_method in inspect.getmembers(target, predicate=inspect.ismethod):
        unbound_method = bound_advisor_method.__func__

        if unbound_method in _advisor_methods:
            for role, name in _advisor_methods[unbound_method]:
                mappings[name].append((role, bound_advisor_method))



"""Test utils.html."""
import pytest

# Websauna
from websauna.utils import html


test_data = (
    ('{"foo": "bar"}', r'{\u0022foo\u0022: \u0022bar\u0022}'),
    ('{"foo": "=-<"}', r'{\u0022foo\u0022: \u0022\u003D\u002D\u003C\u0022}'),
)


@pytest.mark.parametrize('value,expected', test_data)
def test_escape_js(value, expected):
    """Test html.escape_js."""
    func = html.escape_js
    assert func(value) == expected



"""Helpers to traverse trees with __parent__ or __wrapped__ like attributes."""


def traverse_attribute(obj, attribute_name):
    """Turn a recursive parent traverse to a generator.

    E.g. if you have:

    .. code-block:: python

        >>> class Dummy:
        ...     __parent__ = None
        ...
        >>> grandparent = Dummy()
        >>> parent = Dummy()
        >>> parent.__parent__ = grandparent
        >>> obj = Dummy()
        >>> obj.__parent__ = parent

    You can turn this to a list:

    .. code-block:: python

        >>> tree = traverse_attribute(obj, "__parent__")
        >>> list(tree)
        [obj, parent, grandparent]

    """

    while obj is not None:
        yield obj
        obj = getattr(obj, attribute_name, None)



"""Python object fully qualified name resolving."""


def get_qual_name(func: object) -> str:
    """Reverse resolve full dotted name of a Python function.

    Courtesy of http://stackoverflow.com/a/13653312/315168
    """
    return func.__module__ + "." + func.__name__



# Standard Library
import sys
from codecs import open
from os import path

from setuptools import find_packages
from setuptools import setup


assert sys.version_info >= (3, 5, 2), "websauna.utils needs Python 3.5.2 or newer, you have {version}".format(version=sys.version_info)

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='websauna.utils',
    namespace_packages=["websauna"],
    version='1.0a8',
    description=long_description.split("\n")[0],
    long_description=long_description,
    url='https://websauna.org',
    author='Mikko Ohtamaa',
    author_email='mikko@opensourcehacker.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    keywords='config pytest websauna',
    packages=find_packages(exclude=['docs']),
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.5.2',
    install_requires=[

        # User management
        "authomatic",

        # Console logging
        "rainbow_logging_handler",
    ],

    extras_require={
        # Dependencies needed to build and release Websauna
        'dev': [
            'pyroma==2.2',  # This is needed until version 2.4 of Pyroma is released
            'sphinx>=1.6.1',
            'sphinx-autodoc-typehints',
            'sphinx_rtd_theme',
            'sphinxcontrib-zopeext',
            'zest.releaser[recommended]'
        ],
        'test': [
            'codecov',
            'flake8',
            'pytest>=3.0',
            'pytest-runner',
            'coverage',
            'flaky',
            'isort',
            'pytest-cov',
            'pytest-runner',
            'pytest-timeout',
        ],
        "notebook": [
            "ipython[notebook]<5.2",
        ],
        "sqlalchemy": [
            # This package just uses sqlalchemy to enable websauna.utils.jsonb.is_index_property
            "sqlalchemy",
        ]
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'plaster.loader_factory': [
           'ws=websauna.utils.config.loader:Loader',
        ],
        'plaster.wsgi_loader_factory': [
            'ws=websauna.utils.config.loader:Loader',
        ],
    },
)



"""JSONB data utilities."""
# Standard Library
import inspect
import json
from decimal import Decimal

# SQLAlchemy
try:
    from sqlalchemy.ext.indexable import index_property
    _sqlalchemy_present = True
except ImportError:
    _sqlalchemy_present = False


class _DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(o)
        return super(_DecimalEncoder, self).default(o)


def complex_json_dumps(d):
    """Dump JSON so that we handle decimal and dates.

    Decimals are converted to strings.
    """
    return json.dumps(d, cls=_DecimalEncoder)


def is_index_property(obj: object, name: str) -> bool:
    """Check if an object property is index_property like.

    This is needed to correctly generate Colander schema for index_property in SQLAlchemy models.
    If websauna.utils is used in a system without an SQLAlchemy install,
    calling this will raise a RuntimeError.
    """
    # http://docs.sqlalchemy.org/en/rel_1_1/changelog/migration_11.html#new-indexable-orm-extension
    if not _sqlalchemy_present:
        raise RuntimeError("SQLAlchemy not present on this install: can't test ORM's properties")
    attr = inspect.getattr_static(obj, name)
    return isinstance(attr, index_property)


def sanitize_for_json(d: dict) -> dict:
    """Get a JSON round trip for data.

    This ensures we see data as it would be after JSON encode.
    """
    return json.loads(complex_json_dumps(d))



"""Test secrets reader."""
# Standard Library
import os

# Websauna
from websauna.utils import secrets


def test_read_local_secrets():
    """Read file referred by a relative path."""
    data = secrets.read_ini_secrets("websauna/utils/conf/test-secrets.ini", strict=False)
    assert data.get("authomatic.secret") == "xxx"


def test_file_secrets():
    """Read file referred by absolute path."""
    data = secrets.read_ini_secrets("file://" + os.path.abspath("websauna/utils/conf/test-secrets.ini"), strict=False)
    assert data.get("authomatic.secret") == "xxx"
